/*
  Simple Demo for GLSL

  Cleaned up, modified, and re-factored by GZ, June 2007
  Rewritten for Qt5 and OpenGL 3.3 by PD, January 2017
*/

#include <QGuiApplication>
#include <QSurfaceFormat>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QOpenGLWindow>
#include <QOpenGLFunctions_3_3_Core>
#include <QGLFormat>
#include <QtCore/qmath.h>
#include <QtGui/QMatrix4x4>
#include <QtGui/QOpenGLShaderProgram>
#include <QtGui/QScreen>

#include "objloader.hpp"

class Window : public QOpenGLWindow, protected QOpenGLFunctions_3_3_Core
{
public:
	Window();
	~Window();

	void initialize();
	void render();

	void setAnimating(bool animating);
    void setRotating(bool rotating);

public slots:
	void renderLater();
	void renderNow();

protected:
	bool event(QEvent *event) Q_DECL_OVERRIDE;
	void exposeEvent(QExposeEvent *event) Q_DECL_OVERRIDE;
	void resizeGL( int width, int height ) Q_DECL_OVERRIDE;
	void mousePressEvent( QMouseEvent *event ) Q_DECL_OVERRIDE;
	void mouseReleaseEvent( QMouseEvent *event ) Q_DECL_OVERRIDE;
	void mouseMoveEvent( QMouseEvent *event ) Q_DECL_OVERRIDE;
	void keyPressEvent ( QKeyEvent * event ) Q_DECL_OVERRIDE;

private:
	bool m_update_pending;
	bool m_animating;
    bool m_rotating;

	QOpenGLContext * m_context;

	// shader programm id
	GLuint m_program;
	// Vertex Array Object
	GLuint m_vao;
	// data for rendering (vertex position, color)
	GLuint m_data;

	// shader program indices
	GLuint m_posAttr;
	GLuint m_matrixUniform, m_timeUniform;

	// last mouse position for animation
	QPoint m_lastPos;

	// rotation
	float m_xRot, m_yRot, time;

	std::vector<QVector3D> vertices, normals;
	std::vector<QVector2D> uvs;
};



Window::Window()
	: QOpenGLWindow()
	, m_update_pending( false )
	, m_animating( false )
    , m_rotating( true )
	, m_context( 0 )
	, m_program( 0 )
	, m_xRot( 0 )
	, m_yRot( 0 )
	, time( 0 )
{
	setSurfaceType( QWindow::OpenGLSurface );
    setTitle( "Flatten" );
}



Window::~Window()
{
	// cleanup
	glDeleteBuffers( 1, &m_data );
	glDeleteProgram( m_program );
	glDeleteVertexArrays( 1, &m_vao );
}



void Window::initialize()
{
    loadOBJ(":/teapot.obj", vertices, uvs, normals);

	// query OpenGL
	const GLubyte * strTemp = glGetString( GL_RENDERER );
	fprintf(stderr, "GL renderer: %s\n", strTemp ? (const char*)strTemp : "NULL" );

	strTemp = glGetString( GL_VENDOR );
	fprintf(stderr, "GL vendor: %s\n", strTemp ? (const char*)strTemp : "NULL" );

	const GLubyte * strVersion = glGetString( GL_VERSION );
	fprintf(stderr, "GL version: %s\n", strVersion ? (const char*)strVersion : "NULL" );

	qDebug() << "GL context: " << format();

	// set backgroundcolor to black (no red, no green, no blue) with full alpha
	glClearColor( 1, 1, 1, 1 );

	// build shader from files (using Qt)
	{
		// produces memory leak, but is much easier and compatible with Qt resources.
		QOpenGLShaderProgram * program = new QOpenGLShaderProgram();
		program->addShaderFromSourceFile( QOpenGLShader::Vertex, ":/flatten.vert" );
		program->addShaderFromSourceFile( QOpenGLShader::Fragment, ":/flatten.frag" );
		m_program = program->programId();
	}

	// link loaded/compiled vertex and fragment shader
	glLinkProgram( m_program );
	// use compiled program for rendering
	glUseProgram( m_program );

	// get locations indices of attributes and uniforms in activated program to fill them later
	m_posAttr = glGetAttribLocation( m_program, "Vertex" );
	m_matrixUniform = glGetUniformLocation( m_program, "ModelViewProjectionMatrix" );
	m_timeUniform = glGetUniformLocation( m_program, "time" );

	// enable depth test and backface culling
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );

	// Vertex Array Object: stores information about program indices (needed for OpenGL 3+)
	glGenVertexArrays( 1, &m_vao ); // generate 1 buffer
	glBindVertexArray( m_vao ); // use m_data for following commands

	// array buffer containing information about vertex positions, colors
	glGenBuffers( 1, &m_data );
	glBindBuffer( GL_ARRAY_BUFFER, m_data );
	glBufferData( GL_ARRAY_BUFFER, vertices.size() * sizeof(QVector3D), vertices.data(), GL_STATIC_DRAW );

	// Data format: xyzrgb
	glEnableVertexAttribArray( m_posAttr );
	glVertexAttribPointer(
		m_posAttr,           // layout ID in the shader
		3,                   // size (3 * type)
		GL_FLOAT,            // type
		GL_FALSE,            // normalized?
		0,                   // stride
		(void*)0             // array buffer offset (in byte)
	);

	// perspective viewport
	const qreal retinaScale = devicePixelRatio();
	glViewport( 0, 0, width() * retinaScale, height() * retinaScale );
}



void Window::render()
{
	// reset color and depth buffer
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// technically not required here as there is only one object
	glBindVertexArray( m_vao ); // activate the model bound to this vertex array object

	// calculate perspective matrix
	QMatrix4x4 matrix;
	matrix.perspective( 60.0f, 4.0f/3.0f, 1.f, 6.0f );
	matrix.translate( 0, 0, -3 );
	matrix.rotate( 50.0f * m_xRot / screen()->refreshRate(), 0, 1, 0 );
	matrix.rotate( 50.0f * m_yRot / screen()->refreshRate(), 1, 0, 0 );

	// assign matrix to shader
	glUniformMatrix4fv( m_matrixUniform, 1, GL_FALSE, matrix.data() );

	glUniform1f( m_timeUniform, time );

	// draw teapot
	glDrawArrays( GL_TRIANGLES, 0, vertices.size());

	// count frames for animation
    if ( m_rotating )
		++m_xRot;
	time += 0.05f;
}



void Window::renderLater()
{
	if ( !m_update_pending ) {
		m_update_pending = true;
		QCoreApplication::postEvent( this, new QEvent( QEvent::UpdateRequest ) );
	}
}



bool Window::event(QEvent * event)
{
	switch ( event->type() ) {
	case QEvent::UpdateRequest:
		m_update_pending = false;
		renderNow();
		return true;
	default:
		return QWindow::event(event);
	}
}



void Window::exposeEvent( QExposeEvent * )
{
	if ( isExposed() )
		renderNow();
}



void Window::resizeGL( int w, int h )
{
	if ( m_context )
	{
		// perspective viewport
		const qreal retinaScale = devicePixelRatio();
		glViewport( 0, 0, w * retinaScale, h * retinaScale );
	}
}



void Window::mousePressEvent( QMouseEvent * e )
{
	m_lastPos = e->pos();
    setRotating( false );
}



void Window::mouseReleaseEvent( QMouseEvent * )
{
    setRotating( true );
}



void Window::keyPressEvent ( QKeyEvent * e )
{
	if ( !isValid() )
		return;

	switch ( e->key() )
	{
	// close window if Escape was pressed
	case Qt::Key_Escape:
		close();
		break;
	}
}



void Window::mouseMoveEvent( QMouseEvent * e )
{
	int dx = e->x() - m_lastPos.x();
	int dy = e->y() - m_lastPos.y();
	Q_UNUSED(dy);

	// only needed for Mac OS X, but doesn't hurt on other OSes
	bool ctrl_key = e->modifiers() & Qt::MetaModifier;
	Q_UNUSED(ctrl_key);

	if (e->buttons() & Qt::LeftButton)
	{
		m_xRot += dx;
		m_yRot += dy;
	}

	m_lastPos = e->pos();
	e->accept();
	renderLater();
}



void Window::renderNow()
{
	if ( !isExposed() )
		return;

	bool needsInitialize = false;

	if ( !m_context ) {
		m_context = new QOpenGLContext( this );
		m_context->setFormat( requestedFormat() );
		m_context->create();

		needsInitialize = true;
	}

	m_context->makeCurrent( this );

	if ( needsInitialize ) {
		initializeOpenGLFunctions();
		initialize();
	}

	render();

	m_context->swapBuffers(this);

	if ( m_animating )
		renderLater();
}



void Window::setAnimating(bool animating)
{
	m_animating = animating;

	if ( animating )
		renderLater();
}



void Window::setRotating(bool rotating)
{
    m_rotating = rotating;
}



int main(int argc, char **argv)
{
	// configure OpenGL
	QSurfaceFormat format;
	format.setVersion(3, 3); // OpenGL 3.3 Core
	format.setProfile(QSurfaceFormat::CoreProfile);
	format.setSamples(16); // multisampling: anti-aliasing
	format.setDepthBufferSize( 24 );                    // 24 bit for depth buffer
	QSurfaceFormat::setDefaultFormat(format);

    QGuiApplication app(argc, argv);

	// create window and make it visible
	Window window;
	window.setFormat(format);
	window.resize(640, 480);
	window.show();

	// start animation
	window.setAnimating(true);

	return app.exec();
}

